import re


class LDAPBase:
    indent = 4
    collapsed = False
    filters = None

    def simplify(self):
        if self.filters:
            if len(self.filters) == 1:
                return self.filters[0].simplify()
            else:
                self.filters = list(map(lambda x: x.simplify(), self.filters))

        return self

    def to_string(self, indent, level, id_char):
        raise NotImplementedError

    def match(self, data):
        raise NotImplementedError

    @staticmethod
    def _indent(indent, level=0, id_char=' '):
        if type(indent) == bool and indent:
            indent = LDAPBase.indent
        else:
            try:
                indent = int(indent)
            except ValueError:
                return ''

        return id_char * (level * indent)

    @staticmethod
    def escape(data):
        escaped = data.replace('\\', '\\5c')
        escaped = escaped.replace('*', '\\2a')
        escaped = escaped.replace('(', '\\28')
        escaped = escaped.replace(')', '\\29')
        escaped = escaped.replace('\x00', '\\00')

        return escaped

    @staticmethod
    def unescape(data):
        unescaped = data.replace('\\5c', '\\')
        unescaped = unescaped.replace('\\2a', '*')
        unescaped = unescaped.replace('\\28', '(')
        unescaped = unescaped.replace('\\29', ')')
        unescaped = unescaped.replace('\\00', '\x00')

        return unescaped

    @staticmethod
    def match_string(data, filt):
        match = _as_list(data)
        if '*' not in filt:
            return any(_ms_helper(m, filt) for m in match)

        return Filter.match_substring(data, filt)

    @staticmethod
    def match_substring(data, filt):
        match = _as_list(data)

        return any(_ss_helper(m, filt) for m in match)

    @staticmethod
    def match_approx(data, filt):  # TODO: Implement LDAPBase.match_approx()
        pass

    @staticmethod
    def match_lte(data, filt):
        match = _as_list(data)

        return any(_lte_helper(m, filt) for m in match)

    @staticmethod
    def match_gte(data, filt):
        match = _as_list(data)

        return any(_gte_helper(m, filt) for m in match)

    @staticmethod
    def AND(filt):
        return GroupAnd(filt)

    @staticmethod
    def OR(filt):
        return GroupOr(filt)

    @staticmethod
    def NOT(filt):
        filt = _as_list(filt)
        if not len(filt) == 1:  # TODO: Error code here.
            raise Exception

        return GroupNot(filt)


class Filter(LDAPBase):
    def __init__(self, attr, comp, val):
        self.type = 'filter'
        self.attr = attr
        self.comp = comp
        self.val = val

    def __repr__(self):
        return 'Filter: %s' % self.to_string()

    def match(self, data):
        value = self.val

        try:
            attrval = data[self.attr]
        except KeyError:
            return False

        if self.comp == '=':
            if value == '*' and attrval:
                return True
            else:
                return Filter.match_string(attrval, value)
        elif self.comp == '<=':
            return Filter.match_lte(attrval, value)
        elif self.comp == '>=':
            return Filter.match_gte(attrval, value)
        elif self.comp == '~=':
            return Filter.match_approx(attrval, value)
        else:
            pass

    def to_string(self, indent=False, level=0, id_char=' '):
        return ''.join([
            self._indent(indent, level, id_char),
            '(',
            self.attr,
            self.comp,
            self.val,
            ')'
        ])

    @staticmethod
    def attribute(name):
        return Attribute(name)


class Group(LDAPBase):
    def __init__(self, comp, filters):
        self.type = 'group'
        self.comp = comp
        self.filters = filters

    def __repr__(self):
        return 'Filter: %s' % self.to_string()

    def match(self, data):
        raise NotImplementedError

    def to_string(self, indent=False, level=0, id_char=' '):
        id_str = self._indent(indent, level, id_char)
        id_str2 = id_str
        nl = '\n' if indent else ''

        if not Filter.collapsed and self.comp == '!':
            nl = ''
            id_str2 = ''
            indent = 0

        return ''.join([
            id_str,
            '(',
            self.comp,
            nl,
            nl.join(list(map(lambda x: x.to_string(indent, level + 1, id_char), self.filters))),
            nl,
            id_str2,
            ')'
        ])


class GroupOr(Group):
    def __init__(self, filters):
        super().__init__(comp='|', filters=filters)

    def match(self, data):
        return any(f.match(data) for f in self.filters)


class GroupAnd(Group):
    def __init__(self, filters):
        super().__init__(comp='&', filters=filters)

    def match(self, data):
        return any(f.match(data) for f in self.filters)


class GroupNot(Group):
    def __init__(self, filters):
        super().__init__(comp='!', filters=filters)

    def match(self, data):
        return not any(_not_helper(f, data) for f in self.filters)

    def simplify(self):
        return self


class Attribute:
    def __init__(self, name):
        self.name = name

    def present(self):
        return Filter(self.name, '=', '*')

    def raw(self, value):
        return Filter(self.name, '=', value)

    def equal_to(self, value):
        return Filter(self.name, '=', self.escape(value))

    def starts_with(self, value):
        return Filter(self.name, '=', self.escape(value) + '*')

    def ends_with(self, value):
        return Filter(self.name, '=', '*' + self.escape(value))

    def contains(self, value):
        return Filter(self.name, '=', '*' + self.escape(value) + '*')

    def approx(self, value):
        return Filter(self.name, '~=', self.escape(value))

    def lte(self, value):
        return Filter(self.name, '<=', self.escape(value))

    def gte(self, value):
        return Filter(self.name, '>=', self.escape(value))

    @staticmethod
    def escape(data):
        escaped = data.replace('\\', '\\5c')
        escaped = escaped.replace('*', '\\2a')
        escaped = escaped.replace('(', '\\28')
        escaped = escaped.replace(')', '\\29')
        escaped = escaped.replace('\x00', '\\00')

        return escaped


def _as_list(val):
    if not isinstance(val, (list, tuple)):
        return [val]

    return val


def _ss_regex(filt):
    pattern = re.sub(r'\*', '.*', filt)
    pattern = re.sub(r'(?<=\\)([0-9a-fA-F]{,2})', _ss_regex_escaped, pattern)
    return re.compile('^' + pattern + '$', re.I)


def _ss_regex_escaped(match):
    s = match.group(0) if match else None

    if s in ['28', '29', '5c', '2a']:
        s = '\\x{}'.format(match.group(0).upper())

    return s


def _ss_helper(cv, filt):
    regex = _ss_regex(filt)

    return regex.match(cv)


def _ms_helper(cv, filt):
    if cv:
        return cv.lower() == Filter.unescape(filt).lower()


def _lte_helper(cv, filt):
    return cv <= filt


def _gte_helper(cv, filt):
    return cv >= filt


def _not_helper(filt, data):
    try:
        return filt.match(data)
    except AttributeError:
        pass


def _g_to_string_helper(item, indent, level, id_char):
    return item.to_string(indent, level, id_char)



"""
python-ldap-filter
==================

**Build and generate LDAP filters**

A Python utility library for working with Lightweight Directory Access
Protocol (LDAP) filters.

This project is a port of the
`node-ldap-filters <https://github.com/tapmodo/node-ldap-filters>`__ and
implements many of the same APIs. The filters produced by the library
are based on `RFC 4515 <https://tools.ietf.org/html/rfc4515>`__.

Links
-----

`GitHub <https://github.com/SteveEwell/python-ldap-filter>`_

"""

import os
import sys
from setuptools import setup

if sys.version_info[0] <= 2 or (sys.version_info[0] == 3 and sys.version_info < (3, 4)):
    raise RuntimeError('This software requires Python version 3.4 or higher.')

if os.path.isfile("MANIFEST"):
    os.unlink("MANIFEST")

rootdir = os.path.dirname(__file__) or "."


setup(
    name='ldap-filter',
    version='0.1.1',
    description='A Python utility library for working with Lightweight Directory Access Protocol (LDAP) filters.',
    long_description=__doc__,
    url='https://github.com/SteveEwell/python-ldap-filter',
    author='Stephen Ewell',
    author_email='steve@ewell.io',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='ldap filter rfc4515 utility development',
    packages=['ldap_filter'],
    install_requires=[],
    extras_require={
        'test': [
            'pytest>=3.0.2',
            'coverage>=4.2'
        ],
    },
    python_requires='!=2.*,!=3.0.*,!=3.1.*,!=3.2.*,!=3.3.*'
)
